@@ -1,6 +1,6 @@
 /* ----------------------------------------------------------------------- *
  *   
- *   Copyright 1996-2009 The NASM Authors - All Rights Reserved
+ *   Copyright 1996-2016 The NASM Authors - All Rights Reserved
  *   See the file AUTHORS included with the NASM distribution for
  *   the specific copyright holders.
  *
@@ -73,12 +73,14 @@
 #include <time.h>
 #include <stdarg.h>             /* Note: we need the ANSI version of stdarg.h */
 #include <ctype.h>
-#include <inttypes.h>
 
 #include "nasm.h"
 #include "nasmlib.h"
-#include "output/outform.h"
-#include "output/outlib.h"
+#include "error.h"
+#include "ver.h"
+
+#include "outform.h"
+#include "outlib.h"
 
 #ifdef OF_IEEE
 
@@ -139,22 +141,22 @@ static struct ExtBack {
 
 /* NOTE: the first segment MUST be the lineno segment */
 static struct ieeeSection {
-    struct ieeeObjData *data, *datacurr;
     struct ieeeSection *next;
+    char *name;
+    struct ieeeObjData *data, *datacurr;
     struct ieeeFixupp *fptr, *flptr;
     int32_t index;                 /* the NASM segment id */
     int32_t ieee_index;            /* the OBJ-file segment index */
     int32_t currentpos;
     int32_t align;                 /* can be SEG_ABS + absolute addr */
     int32_t startpos;
+    int32_t use32;                 /* is this segment 32-bit? */
+    struct ieeePublic *pubhead, **pubtail, *lochead, **loctail;
     enum {
         CMB_PRIVATE = 0,
         CMB_PUBLIC = 2,
         CMB_COMMON = 6
     } combine;
-    int32_t use32;                 /* is this segment 32-bit? */
-    struct ieeePublic *pubhead, **pubtail, *lochead, **loctail;
-    char *name;
 } *seghead, **segtail, *ieee_seg_needs_update;
 
 struct ieeeObjData {
@@ -184,14 +186,15 @@ struct ieeeFixupp {
 static int32_t ieee_entry_seg, ieee_entry_ofs;
 static int checksum;
 
-extern struct ofmt of_ieee;
+extern const struct ofmt of_ieee;
+static const struct dfmt ladsoft_debug_form;
 
 static void ieee_data_new(struct ieeeSection *);
 static void ieee_write_fixup(int32_t, int32_t, struct ieeeSection *,
                              int, uint64_t, int32_t);
 static void ieee_install_fixup(struct ieeeSection *, struct ieeeFixupp *);
 static int32_t ieee_segment(char *, int, int *);
-static void ieee_write_file(int debuginfo);
+static void ieee_write_file(void);
 static void ieee_write_byte(struct ieeeSection *, int);
 static void ieee_write_word(struct ieeeSection *, int);
 static void ieee_write_dword(struct ieeeSection *, int32_t);
@@ -221,21 +224,13 @@ static void ieee_init(void)
     checksum = 0;
 }
 
-static int ieee_set_info(enum geninfo type, char **val)
-{
-    (void)type;
-    (void)val;
-
-    return 0;
-}
-
 /*
  * Rundown
  */
-static void ieee_cleanup(int debuginfo)
+static void ieee_cleanup(void)
 {
-    ieee_write_file(debuginfo);
-    of_ieee.current_dfmt->cleanup();
+    ieee_write_file();
+    dfmt->cleanup();
     while (seghead) {
         struct ieeeSection *segtmp = seghead;
         seghead = seghead->next;
@@ -305,7 +300,7 @@ static void ieee_deflabel(char *name, int32_t segment,
      * First check for the double-period, signifying something
      * unusual.
      */
-    if (name[0] == '.' && name[1] == '.') {
+    if (name[0] == '.' && name[1] == '.' && name[2] != '@') {
         if (!strcmp(name, "..start")) {
             ieee_entry_seg = segment;
             ieee_entry_ofs = offset;
@@ -419,7 +414,7 @@ static void ieee_out(int32_t segto, const void *data,
     if (!any_segs) {
         int tempint;            /* ignored */
         if (segto != ieee_segment("__NASMDEFSEG", 2, &tempint))
-            nasm_error(ERR_PANIC, "strange segment conditions in IEEE driver");
+            nasm_panic(0, "strange segment conditions in IEEE driver");
     }
 
     /*
@@ -429,7 +424,7 @@ static void ieee_out(int32_t segto, const void *data,
         if (seg->index == segto)
             break;
     if (!seg)
-        nasm_error(ERR_PANIC, "code directed to nonexistent segment?");
+        nasm_panic(0, "code directed to nonexistent segment?");
 
     if (type == OUT_RAWDATA) {
         ucdata = data;
@@ -437,7 +432,9 @@ static void ieee_out(int32_t segto, const void *data,
             ieee_write_byte(seg, *ucdata++);
     } else if (type == OUT_ADDRESS || type == OUT_REL2ADR ||
                type == OUT_REL4ADR) {
-        if (segment == NO_SEG && type != OUT_ADDRESS)
+        if (type == OUT_ADDRESS)
+            size = abs((int)size);
+        else if (segment == NO_SEG)
             nasm_error(ERR_NONFATAL, "relative call to absolute address not"
                   " supported by IEEE format");
         ldata = *(int64_t *)data;
@@ -533,7 +530,7 @@ static void ieee_write_fixup(int32_t segment, int32_t wrt,
                     }
 
                 } else
-                    nasm_error(ERR_PANIC,
+                    nasm_panic(0,
                           "unrecognised WRT value in ieee_write_fixup");
             } else
                 nasm_error(ERR_NONFATAL, "target of WRT must be a section ");
@@ -575,7 +572,7 @@ static void ieee_write_fixup(int32_t segment, int32_t wrt,
                  */
                 if (eb) {
                     if (realtype == OUT_REL2ADR || realtype == OUT_REL4ADR) {
-                        nasm_error(ERR_PANIC,
+                        nasm_panic(0,
                               "Segment of a rel not supported in ieee_write_fixup");
                     } else {
                         /* If we want the segment */
@@ -586,7 +583,7 @@ static void ieee_write_fixup(int32_t segment, int32_t wrt,
 
                 } else
                     /* If we get here the seg value doesn't make sense */
-                    nasm_error(ERR_PANIC,
+                    nasm_panic(0,
                           "unrecognised segment value in ieee_write_fixup");
             }
 
@@ -641,7 +638,7 @@ static void ieee_write_fixup(int32_t segment, int32_t wrt,
 
                 } else
                     /* If we get here the seg value doesn't make sense */
-                    nasm_error(ERR_PANIC,
+                    nasm_panic(0,
                           "unrecognised segment value in ieee_write_fixup");
             }
         }
@@ -783,7 +780,7 @@ static int32_t ieee_segment(char *name, int pass, int *bits)
                     nasm_error(ERR_NONFATAL, "segment alignment should be"
                           " numeric");
                 }
-                switch ((int)seg->align) {
+                switch (seg->align) {
                 case 1:        /* BYTE */
                 case 2:        /* WORD */
                 case 4:        /* DWORD */
@@ -827,7 +824,8 @@ static int32_t ieee_segment(char *name, int pass, int *bits)
 /*
  * directives supported
  */
-static int ieee_directive(enum directives directive, char *value, int pass)
+static enum directive_result
+ieee_directive(enum directive directive, char *value, int pass)
 {
 
     (void)value;
@@ -836,10 +834,10 @@ static int ieee_directive(enum directives directive, char *value, int pass)
     switch (directive) {
     case D_UPPERCASE:
         ieee_uppercase = true;
-        return 1;
+        return DIRR_OK;
 	
     default:
-	return 0;
+	return DIRR_UNKNOWN;
     }
 }
 
@@ -897,7 +895,7 @@ static void ieee_filename(char *inname, char *outname)
     standard_extension(inname, outname, ".o");
 }
 
-static void ieee_write_file(int debuginfo)
+static void ieee_write_file(void)
 {
     struct tm *thetime;
     time_t reltime;
@@ -909,6 +907,7 @@ static void ieee_write_file(int debuginfo)
     struct ieeeFixupp *fix;
     struct Array *arr;
     int i;
+    const bool debuginfo = (dfmt == &ladsoft_debug_form);
 
     /*
      * Write the module header
@@ -992,7 +991,7 @@ static void ieee_write_file(int debuginfo)
             if (seg->index == ieee_entry_seg)
                 break;
         if (!seg)
-            nasm_error(ERR_PANIC, "Start address records are incorrect");
+            nasm_panic(0, "Start address records are incorrect");
         else
             ieee_putascii("ASG,R%X,%lX,+.\n", seg->ieee_index,
                           ieee_entry_ofs);
@@ -1297,7 +1296,7 @@ static void ieee_unqualified_name(char *dest, char *source)
     } else
         strcpy(dest, source);
 }
-void dbgls_init(void)
+static void dbgls_init(void)
 {
     int tempint;
 
@@ -1355,7 +1354,7 @@ static void dbgls_linnum(const char *lnfname, int32_t lineno, int32_t segto)
     if (!any_segs) {
         int tempint;            /* ignored */
         if (segto != ieee_segment("__NASMDEFSEG", 2, &tempint))
-            nasm_error(ERR_PANIC, "strange segment conditions in OBJ driver");
+            nasm_panic(0, "strange segment conditions in OBJ driver");
     }
 
     /*
@@ -1365,7 +1364,7 @@ static void dbgls_linnum(const char *lnfname, int32_t lineno, int32_t segto)
         if (seg->index == segto)
             break;
     if (!seg)
-        nasm_error(ERR_PANIC, "lineno directed to nonexistent segment?");
+        nasm_panic(0, "lineno directed to nonexistent segment?");
 
     for (fn = fnhead; fn; fn = fn->next) {
         if (!nasm_stricmp(lnfname, fn->name))
@@ -1395,20 +1394,16 @@ static void dbgls_deflabel(char *name, int32_t segment,
     /* Keep compiler from warning about special */
     (void)special;
 
+    /*
+     * Note: ..[^@] special symbols are filtered in labels.c
+     */
+
     /*
      * If it's a special-retry from pass two, discard it.
      */
     if (is_global == 3)
         return;
 
-    /*
-     * First check for the double-period, signifying something
-     * unusual.
-     */
-    if (name[0] == '.' && name[1] == '.' && name[2] != '@') {
-        return;
-    }
-
     /*
      * Case (i):
      */
@@ -1495,7 +1490,7 @@ static void dbgls_output(int output_type, void *param)
     (void)output_type;
     (void)param;
 }
-static struct dfmt ladsoft_debug_form = {
+static const struct dfmt ladsoft_debug_form = {
     "LADsoft Debug Records",
     "ladsoft",
     dbgls_init,
@@ -1505,21 +1500,23 @@ static struct dfmt ladsoft_debug_form = {
     dbgls_typevalue,
     dbgls_output,
     dbgls_cleanup,
+    NULL                        /* pragma list */
 };
-static struct dfmt *ladsoft_debug_arr[3] = {
+static const struct dfmt * const ladsoft_debug_arr[3] = {
     &ladsoft_debug_form,
     &null_debug_form,
     NULL
 };
-struct ofmt of_ieee = {
+const struct ofmt of_ieee = {
     "IEEE-695 (LADsoft variant) object file format",
     "ieee",
     OFMT_TEXT,
+    32,
     ladsoft_debug_arr,
     &ladsoft_debug_form,
     NULL,
     ieee_init,
-    ieee_set_info,
+    nasm_do_legacy_output,
     ieee_out,
     ieee_deflabel,
     ieee_segment,
@@ -1527,7 +1524,8 @@ struct ofmt of_ieee = {
     ieee_segbase,
     ieee_directive,
     ieee_filename,
-    ieee_cleanup
+    ieee_cleanup,
+    NULL                        /* pragma list */
 };
 
 #endif                          /* OF_IEEE */
