@@ -1,6 +1,6 @@
 /* ----------------------------------------------------------------------- *
  *
- *   Copyright 1996-2016 The NASM Authors - All Rights Reserved
+ *   Copyright 1996-2017 The NASM Authors - All Rights Reserved
  *   See the file AUTHORS included with the NASM distribution for
  *   the specific copyright holders.
  *
@@ -42,15 +42,18 @@
 #include <stdlib.h>
 #include <string.h>
 #include <ctype.h>
-#include <inttypes.h>
 
 #include "nasm.h"
 #include "nasmlib.h"
+#include "labels.h"
+#include "error.h"
 #include "saa.h"
 #include "raa.h"
 #include "rbtree.h"
-#include "output/outform.h"
-#include "output/outlib.h"
+#include "outform.h"
+#include "outlib.h"
+#include "ver.h"
+#include "dwarf.h"
 
 #if defined(OF_MACHO) || defined(OF_MACHO64)
 
@@ -75,6 +78,9 @@
 #define	CPU_SUBTYPE_I386_ALL	3		/* all-x86 compatible */
 #define	MH_OBJECT		0x1		/* object file */
 
+/* Mach-O header flags */
+#define MH_SUBSECTIONS_VIA_SYMBOLS 0x2000
+
 /* Mach-O load commands */
 #define LC_SEGMENT		0x1		/* 32-bit segment load cmd */
 #define LC_SEGMENT_64		0x19		/* 64-bit segment load cmd */
@@ -166,35 +172,28 @@ struct section {
 
 #define SECTION_TYPE	0x000000ff      /* section type mask */
 
-#define	S_REGULAR	(0x0)   /* standard section */
-#define	S_ZEROFILL	(0x1)   /* zerofill, in-memory only */
+#define	S_REGULAR		(0x0)   /* standard section */
+#define	S_ZEROFILL		(0x1)   /* zerofill, in-memory only */
 
 #define SECTION_ATTRIBUTES_SYS   0x00ffff00     /* system setable attributes */
 #define S_ATTR_SOME_INSTRUCTIONS 0x00000400     /* section contains some
-                                                   machine instructions */
-#define S_ATTR_EXT_RELOC         0x00000200     /* section has external
-                                                   relocation entries */
-#define S_ATTR_LOC_RELOC         0x00000100     /* section has local
-                                                   relocation entries */
-#define S_ATTR_PURE_INSTRUCTIONS 0x80000000		/* section uses pure
-												   machine instructions */
+						   machine instructions */
+#define S_ATTR_EXT_RELOC         0x00000200     /* section has external relocation entries */
+#define S_ATTR_LOC_RELOC         0x00000100     /* section has local relocation entries */
+#define S_ATTR_DEBUG		 0x02000000
+#define S_ATTR_SELF_MODIFYING_CODE 0x04000000
+#define S_ATTR_LIVE_SUPPORT	 0x08000000
+#define S_ATTR_NO_DEAD_STRIP	 0x10000000     /* no dead stripping */
+#define S_ATTR_STRIP_STATIC_SYMS 0x20000000
+#define S_ATTR_NO_TOC            0x40000000
+#define S_ATTR_PURE_INSTRUCTIONS 0x80000000	/* section uses pure machine instructions */
+#define S_ATTR_DEBUG 0x02000000 /* debug section */
 
-/* Fake section for absolute symbols, *not* part of the section linked list */
+#define S_NASM_TYPE_MASK	 0x800004ff	/* we consider these bits "section type" */
+
+/* fake section for absolute symbols, *not* part of the section linked list */
 static struct section absolute_sect;
 
-static const struct sectmap {
-    const char *nasmsect;
-    const char *segname;
-    const char *sectname;
-    const int32_t flags;
-} sectmap[] = {
-    {".text", "__TEXT", "__text", S_REGULAR|S_ATTR_SOME_INSTRUCTIONS|S_ATTR_PURE_INSTRUCTIONS},
-    {".data", "__DATA", "__data", S_REGULAR},
-    {".rodata", "__DATA", "__const", S_REGULAR},
-    {".bss", "__DATA", "__bss", S_ZEROFILL},
-    {NULL, NULL, NULL, 0}
-};
-
 struct reloc {
     /* nasm internal data */
     struct reloc *next;
@@ -281,6 +280,7 @@ static uint32_t strslen;
    a structure or as function arguments.  */
 static uint32_t head_ncmds = 0;
 static uint32_t head_sizeofcmds = 0;
+static uint32_t head_flags = 0;
 static uint64_t seg_filesize = 0;
 static uint64_t seg_vmsize = 0;
 static uint32_t seg_nsects = 0;
@@ -323,6 +323,49 @@ static struct section *get_section_by_index(const int32_t index)
     return s;
 }
 
+struct dir_list {
+    struct dir_list *next;
+    struct dir_list *last;
+    const char *dir_name;
+    uint32_t dir;
+};
+
+struct file_list {
+    struct file_list *next;
+    struct file_list *last;
+    const char *file_name;
+    uint32_t file;
+    struct dir_list *dir;
+};
+
+struct dw_sect_list {
+    struct SAA *psaa;
+    int32_t section;
+    uint32_t line;
+    uint64_t offset;
+    uint32_t file;
+    struct dw_sect_list *next;
+    struct dw_sect_list *last;
+};
+
+struct section_info {
+    uint64_t size;
+    int32_t secto;
+};
+
+#define DW_LN_BASE (-5)
+#define DW_LN_RANGE 14
+#define DW_OPCODE_BASE 13
+#define DW_MAX_LN (DW_LN_BASE + DW_LN_RANGE)
+#define DW_MAX_SP_OPCODE 256
+
+static struct file_list *dw_head_file = 0, *dw_cur_file = 0, **dw_last_file_next = NULL;
+static struct dir_list *dw_head_dir = 0, **dw_last_dir_next = NULL;
+static struct dw_sect_list  *dw_head_sect = 0, *dw_cur_sect = 0, *dw_last_sect = 0;
+static uint32_t  cur_line = 0, dw_num_files = 0, dw_num_dirs = 0, dw_num_sects = 0;
+static bool  dbg_immcall = false;
+static const char *module_name = NULL;
+
 /*
  * Special section numbers which are used to define Mach-O special
  * symbols, which can be used with WRT to provide PIC relocation
@@ -554,6 +597,16 @@ static void macho_output(int32_t secto, const void *data,
             nasm_panic(0, "text section not found");
     }
 
+    /* debug code generation only for sections tagged with
+     * instruction attribute */
+    if (s->flags & S_ATTR_SOME_INSTRUCTIONS)
+    {
+        struct section_info sinfo;
+        sinfo.size = s->size;
+        sinfo.secto = secto;
+        dfmt->debug_output(0, &sinfo);
+    }
+
     is_bss = (s->flags & SECTION_TYPE) == S_ZEROFILL;
 
     if (is_bss && type != OUT_RESERVE) {
@@ -704,6 +757,43 @@ static void macho_output(int32_t secto, const void *data,
     }
 }
 
+/* Translation table from traditional Unix section names to Mach-O */
+static const struct sectmap {
+    const char *nasmsect;
+    const char *segname;
+    const char *sectname;
+    const uint32_t flags;
+} sectmap[] = {
+    {".text", "__TEXT", "__text",
+     S_REGULAR|S_ATTR_SOME_INSTRUCTIONS|S_ATTR_PURE_INSTRUCTIONS},
+    {".data", "__DATA", "__data", S_REGULAR},
+    {".rodata", "__DATA", "__const", S_REGULAR},
+    {".bss", "__DATA", "__bss", S_ZEROFILL},
+    {".debug_abbrev", "__DWARF", "__debug_abbrev", S_ATTR_DEBUG},
+    {".debug_info", "__DWARF", "__debug_info", S_ATTR_DEBUG},
+    {".debug_line", "__DWARF", "__debug_line", S_ATTR_DEBUG},
+    {".debug_str", "__DWARF", "__debug_str", S_ATTR_DEBUG},
+    {NULL, NULL, NULL, 0}
+};
+
+#define NO_TYPE S_NASM_TYPE_MASK
+
+/* Section type or attribute directives */
+static const struct sect_attribs {
+    const char *name;
+    uint32_t flags;
+} sect_attribs[] = {
+    { "data", S_REGULAR },
+    { "code", S_REGULAR|S_ATTR_SOME_INSTRUCTIONS|S_ATTR_PURE_INSTRUCTIONS },
+    { "mixed", S_REGULAR|S_ATTR_SOME_INSTRUCTIONS },
+    { "bss", S_ZEROFILL },
+    { "zerofill", S_ZEROFILL },
+    { "no_dead_strip", NO_TYPE|S_ATTR_NO_DEAD_STRIP },
+    { "live_support", NO_TYPE|S_ATTR_LIVE_SUPPORT },
+    { "strip_static_syms", NO_TYPE|S_ATTR_STRIP_STATIC_SYMS },
+    { NULL, 0 }
+};
+
 static int32_t macho_section(char *name, int pass, int *bits)
 {
     char *sectionAttributes;
@@ -711,8 +801,10 @@ static int32_t macho_section(char *name, int pass, int *bits)
     struct section *s;
     const char *section, *segment;
     uint32_t flags;
+    const struct sect_attribs *sa;
     char *currentAttribute;
     char *comma;
+
     bool new_seg;
 
     (void)pass;
@@ -807,61 +899,66 @@ static int32_t macho_section(char *name, int pass, int *bits)
 
     s->by_name = s->by_name || comma; /* Was specified by name */
 
-    flags = (uint32_t)-1;
+    flags = NO_TYPE;
 
-    while ((NULL != sectionAttributes)
-	   && (currentAttribute = nasm_strsep(&sectionAttributes, " \t"))) {
-	if (0 != *currentAttribute) {
-	    if (!nasm_strnicmp("align=", currentAttribute, 6)) {
-		char *end;
-		int newAlignment, value;
+    while (sectionAttributes &&
+	   (currentAttribute = nasm_strsep(&sectionAttributes, " \t"))) {
+	if (!*currentAttribute)
+	    continue;
 
-		value = strtoul(currentAttribute + 6, (char**)&end, 0);
-		newAlignment = alignlog2_32(value);
+	if (!nasm_strnicmp("align=", currentAttribute, 6)) {
+	    char *end;
+	    int newAlignment, value;
 
-		if (0 != *end) {
-		    nasm_error(ERR_NONFATAL,
-			       "unknown or missing alignment value \"%s\" "
-			       "specified for section \"%s\"",
-			       currentAttribute + 6,
-			       name);
-		} else if (0 > newAlignment) {
-		    nasm_error(ERR_NONFATAL,
-			       "alignment of %d (for section \"%s\") is not "
-			       "a power of two",
-			       value,
-			       name);
+	    value = strtoul(currentAttribute + 6, (char**)&end, 0);
+	    newAlignment = alignlog2_32(value);
+
+	    if (0 != *end) {
+		nasm_error(ERR_NONFATAL,
+			   "unknown or missing alignment value \"%s\" "
+			   "specified for section \"%s\"",
+			   currentAttribute + 6,
+			   name);
+	    } else if (0 > newAlignment) {
+		nasm_error(ERR_NONFATAL,
+			   "alignment of %d (for section \"%s\") is not "
+			   "a power of two",
+			   value,
+			   name);
+	    }
+
+	    if (s->align < newAlignment)
+		s->align = newAlignment;
+	} else {
+	    for (sa = sect_attribs; sa->name; sa++) {
+		if (!nasm_stricmp(sa->name, currentAttribute)) {
+		    if ((sa->flags & S_NASM_TYPE_MASK) != NO_TYPE) {
+			flags = (flags & ~S_NASM_TYPE_MASK)
+			    | (sa->flags & S_NASM_TYPE_MASK);
+		    }
+		    flags |= sa->flags & ~S_NASM_TYPE_MASK;
+		    break;
 		}
+	    }
 
-		if (s->align < newAlignment)
-		    s->align = newAlignment;
-	    } else if (!nasm_stricmp("data", currentAttribute)) {
-		flags = S_REGULAR;
-	    } else if (!nasm_stricmp("code", currentAttribute) ||
-		       !nasm_stricmp("text", currentAttribute)) {
-		flags = S_REGULAR | S_ATTR_SOME_INSTRUCTIONS |
-		    S_ATTR_PURE_INSTRUCTIONS;
-	    } else if (!nasm_stricmp("mixed", currentAttribute)) {
-		flags = S_REGULAR | S_ATTR_SOME_INSTRUCTIONS;
-	    } else if (!nasm_stricmp("bss", currentAttribute)) {
-		flags = S_ZEROFILL;
-	    } else {
+	    if (!sa->name) {
 		nasm_error(ERR_NONFATAL,
 			   "unknown section attribute %s for section %s",
-			   currentAttribute,
-			   name);
+			   currentAttribute, name);
 	    }
 	}
+    }
 
-	if (flags != (uint32_t)-1) {
-	    if (!new_seg && s->flags != flags) {
-		nasm_error(ERR_NONFATAL,
-			   "inconsistent section attributes for section %s\n",
-			   name);
-	    } else {
-		s->flags = flags;
-	    }
+    if ((flags & S_NASM_TYPE_MASK) != NO_TYPE) {
+	if (!new_seg && ((s->flags ^ flags) & S_NASM_TYPE_MASK)) {
+	    nasm_error(ERR_NONFATAL,
+		       "inconsistent section attributes for section %s\n",
+		       name);
+	} else {
+	    s->flags = (s->flags & ~S_NASM_TYPE_MASK) | flags;
 	}
+    } else {
+	s->flags |= flags & ~S_NASM_TYPE_MASK;
     }
 
     return s->index;
@@ -984,6 +1081,7 @@ static int32_t macho_segbase(int32_t section)
 static void macho_filename(char *inname, char *outname)
 {
     standard_extension(inname, outname, ".o");
+    module_name = inname;
 }
 
 extern macros_t macho_stdmac[];
@@ -1028,11 +1126,11 @@ static void macho_layout_symbols (uint32_t *numsyms,
 	    nlocalsym++;
 	}
 	else {
-		if ((sym->type & N_TYPE) != N_UNDF) {
-			nextdefsym++;
+	    if ((sym->type & N_TYPE) != N_UNDF) {
+		nextdefsym++;
 	    } else {
-			nundefsym++;
-		}
+		nundefsym++;
+	    }
 
 	    /* If we handle debug info we'll want
 	       to check for it here instead of just
@@ -1068,11 +1166,11 @@ static void macho_layout_symbols (uint32_t *numsyms,
 	    *strtabsize += strlen(sym->name) + 1;
 	}
 	else {
-		if((sym->type & N_TYPE) != N_UNDF) {
-			extdefsyms[i++] = sym;
+	    if ((sym->type & N_TYPE) != N_UNDF) {
+		extdefsyms[i++] = sym;
 	    } else {
-			undefsyms[j++] = sym;
-		}
+		undefsyms[j++] = sym;
+	    }
 	}
 	symp = &(sym->next);
     }
@@ -1110,7 +1208,7 @@ static void macho_calculate_sizes (void)
         if (s->align == -1)
             s->align = DEFAULT_SECTION_ALIGNMENT;
 
-	newaddr = ALIGN(s->addr, 1 << s->align);
+        newaddr = ALIGN(s->addr, UINT64_C(1) << s->align);
         s->addr = newaddr;
 
         seg_vmsize = newaddr + s->size;
@@ -1125,6 +1223,9 @@ static void macho_calculate_sizes (void)
 	    s->pad = ALIGN(seg_filesize, 4) - seg_filesize;
 	    s->offset = seg_filesize + s->pad;
             seg_filesize += s->size + s->pad;
+
+            /* filesize and vmsize needs to be aligned */
+            seg_vmsize += s->pad;
 	}
     }
 
@@ -1162,7 +1263,7 @@ static void macho_write_header (void)
     fwriteint32_t(MH_OBJECT, ofile);	/* Mach-O file type */
     fwriteint32_t(head_ncmds, ofile);	/* number of load commands */
     fwriteint32_t(head_sizeofcmds, ofile);	/* size of load commands */
-    fwriteint32_t(0, ofile);			/* no flags */
+    fwriteint32_t(head_flags, ofile);		/* flags, if any */
     fwritezero(fmt.header_size - 7*4, ofile);	/* reserved fields */
 }
 
@@ -1553,6 +1654,8 @@ static void macho_cleanup(void)
     struct reloc *r;
     struct symbol *sym;
 
+    dfmt->cleanup();
+
     /* Sort all symbols.  */
     macho_layout_symbols (&nsyms, &strslen);
 
@@ -1594,6 +1697,528 @@ static void macho_cleanup(void)
     nasm_free(sectstab);
 }
 
+static bool macho_set_section_attribute_by_symbol(const char *label, uint32_t flags)
+{
+    struct section *s;
+    int32_t nasm_seg;
+    int64_t offset;
+
+    if (!lookup_label(label, &nasm_seg, &offset)) {
+	nasm_error(ERR_NONFATAL, "unknown symbol `%s' in no_dead_strip", label);
+	return false;
+    }
+
+    s = get_section_by_index(nasm_seg);
+    if (!s) {
+	nasm_error(ERR_NONFATAL, "symbol `%s' is external or absolute", label);
+	return false;
+    }
+
+    s->flags |= flags;
+    return true;
+}
+
+/*
+ * Mark a symbol for no dead stripping
+ */
+static enum directive_result macho_no_dead_strip(const char *labels)
+{
+    char *s, *p, *ep;
+    char ec;
+    enum directive_result rv = DIRR_ERROR;
+    bool real = passn > 1;
+
+    p = s = nasm_strdup(labels);
+    while (*p) {
+	ep = nasm_skip_identifier(p);
+	if (!ep) {
+	    nasm_error(ERR_NONFATAL, "invalid symbol in NO_DEAD_STRIP");
+	    goto err;
+	}
+	ec = *ep;
+	if (ec && ec != ',' && !nasm_isspace(ec)) {
+	    nasm_error(ERR_NONFATAL, "cannot parse contents after symbol");
+	    goto err;
+	}
+	*ep = '\0';
+	if (real) {
+	    if (!macho_set_section_attribute_by_symbol(p, S_ATTR_NO_DEAD_STRIP))
+		rv = DIRR_ERROR;
+	}
+	*ep = ec;
+	p = nasm_skip_spaces(ep);
+	if (*p == ',')
+	    p = nasm_skip_spaces(++p);
+    }
+
+    rv = DIRR_OK;
+
+err:
+    nasm_free(s);
+    return rv;
+}
+
+/*
+ * Mach-O pragmas
+ */
+static enum directive_result
+macho_pragma(const struct pragma *pragma)
+{
+    bool real = passn > 1;
+
+    switch (pragma->opcode) {
+    case D_SUBSECTIONS_VIA_SYMBOLS:
+	if (*pragma->tail)
+	    return DIRR_BADPARAM;
+
+	if (real)
+	    head_flags |= MH_SUBSECTIONS_VIA_SYMBOLS;
+
+	return DIRR_OK;
+
+    case D_NO_DEAD_STRIP:
+	return macho_no_dead_strip(pragma->tail);
+
+    default:
+	return DIRR_UNKNOWN;	/* Not a Mach-O directive */
+    }
+}
+
+static const struct pragma_facility macho_pragma_list[] = {
+    { "macho", macho_pragma },
+    { NULL, macho_pragma }	/* Implements macho32/macho64 namespaces */
+};
+
+static void macho_dbg_generate(void)
+{
+    uint8_t *p_buf = NULL, *p_buf_base = NULL;
+    size_t saa_len = 0, high_addr = 0, total_len = 0;
+    struct section *p_section = NULL;
+    /* calculated at debug_str and referenced at debug_info */
+    uint32_t producer_str_offset = 0, module_str_offset = 0, dir_str_offset = 0;
+
+    /* debug section defines */
+    {
+        int bits = 0;
+        macho_section(".debug_abbrev", 0, &bits);
+        macho_section(".debug_info", 0, &bits);
+        macho_section(".debug_line", 0, &bits);
+        macho_section(".debug_str", 0, &bits);
+    }
+
+    /* dw section walk to find high_addr and total_len */
+    {
+        struct dw_sect_list *p_sect;
+
+        list_for_each(p_sect, dw_head_sect) {
+            uint64_t offset = get_section_by_index(p_sect->section)->size;
+            struct SAA *p_linep = p_sect->psaa;
+
+            saa_write8(p_linep, 2); /* std op 2 */
+            saa_write8(p_linep, offset - p_sect->offset);
+            saa_write8(p_linep, DW_LNS_extended_op);
+            saa_write8(p_linep, 1); /* operand length */
+            saa_write8(p_linep, DW_LNE_end_sequence);
+
+            total_len += p_linep->datalen;
+            high_addr += offset;
+        }
+    }
+
+    /* debug line */
+    {
+        struct dw_sect_list *p_sect;
+        size_t linep_off, buf_size;
+        struct SAA *p_lines = saa_init(1L);
+        struct dir_list *p_dir;
+        struct file_list *p_file;
+
+        p_section = get_section_by_name("__DWARF", "__debug_line");
+        nasm_assert(p_section != NULL);
+
+        saa_write8(p_lines, 1); /* minimum instruction length */
+        saa_write8(p_lines, 1); /* initial value of "is_stmt" */
+        saa_write8(p_lines, DW_LN_BASE); /* line base */
+        saa_write8(p_lines, DW_LN_RANGE); /* line range */
+        saa_write8(p_lines, DW_OPCODE_BASE); /* opcode base */
+        saa_write8(p_lines, 0); /* std opcode 1 length */
+        saa_write8(p_lines, 1); /* std opcode 2 length */
+        saa_write8(p_lines, 1); /* std opcode 3 length */
+        saa_write8(p_lines, 1); /* std opcode 4 length */
+        saa_write8(p_lines, 1); /* std opcode 5 length */
+        saa_write8(p_lines, 0); /* std opcode 6 length */
+        saa_write8(p_lines, 0); /* std opcode 7 length */
+        saa_write8(p_lines, 0); /* std opcode 8 length */
+        saa_write8(p_lines, 1); /* std opcode 9 length */
+        saa_write8(p_lines, 0); /* std opcode 10 length */
+        saa_write8(p_lines, 0); /* std opcode 11 length */
+        saa_write8(p_lines, 1); /* std opcode 12 length */
+        list_for_each(p_dir, dw_head_dir) {
+            saa_wcstring(p_lines, p_dir->dir_name);
+        }
+        saa_write8(p_lines, 0); /* end of table */
+
+        list_for_each(p_file, dw_head_file) {
+            saa_wcstring(p_lines, p_file->file_name);
+            saa_write8(p_lines, p_file->dir->dir); /* directory id */
+            saa_write8(p_lines, 0); /* time */
+            saa_write8(p_lines, 0); /* size */
+        }
+        saa_write8(p_lines, 0); /* end of table */
+
+        linep_off = p_lines->datalen;
+        /* 10 bytes for initial & prolong length, and dwarf version info */
+        buf_size = saa_len = linep_off + total_len + 10;
+        p_buf_base = p_buf = nasm_malloc(buf_size);
+
+        WRITELONG(p_buf, saa_len - 4); /* initial length; size excluding itself */
+        WRITESHORT(p_buf, 2); /* dwarf version */
+        WRITELONG(p_buf, linep_off); /* prolong length */
+
+        saa_rnbytes(p_lines, p_buf, linep_off);
+        p_buf += linep_off;
+        saa_free(p_lines);
+
+        list_for_each(p_sect, dw_head_sect) {
+            struct SAA *p_linep = p_sect->psaa;
+
+            saa_len = p_linep->datalen;
+            saa_rnbytes(p_linep, p_buf, saa_len);
+            p_buf += saa_len;
+
+            saa_free(p_linep);
+        }
+
+        macho_output(p_section->index, p_buf_base, OUT_RAWDATA, buf_size, NO_SEG, 0);
+
+        nasm_free(p_buf_base);
+    }
+
+    /* string section */
+    {
+        struct SAA *p_str = saa_init(1L);
+        char *cur_path = nasm_realpath(module_name);
+        char *cur_file = nasm_basename(cur_path);
+        char *cur_dir = nasm_dirname(cur_path);
+
+        p_section = get_section_by_name("__DWARF", "__debug_str");
+        nasm_assert(p_section != NULL);
+
+        producer_str_offset = 0;
+        module_str_offset = dir_str_offset = saa_wcstring(p_str, nasm_signature);
+        dir_str_offset += saa_wcstring(p_str, cur_file);
+        saa_wcstring(p_str, cur_dir);
+
+        saa_len = p_str->datalen;
+        p_buf = nasm_malloc(saa_len);
+        saa_rnbytes(p_str, p_buf, saa_len);
+        macho_output(p_section->index, p_buf, OUT_RAWDATA, saa_len, NO_SEG, 0);
+
+        nasm_free(cur_path);
+        nasm_free(cur_file);
+        nasm_free(cur_dir);
+        saa_free(p_str);
+        nasm_free(p_buf);
+    }
+
+    /* debug info */
+    {
+        struct SAA *p_info = saa_init(1L);
+
+        p_section = get_section_by_name("__DWARF", "__debug_info");
+        nasm_assert(p_section != NULL);
+
+        /* size will be overwritten once determined, so skip in p_info layout */
+        saa_write16(p_info, 2); /* dwarf version */
+        saa_write32(p_info, 0); /* offset info abbrev */
+        saa_write8(p_info, (ofmt == &of_macho64) ? 8 : 4);   /* pointer size  */
+
+        saa_write8(p_info, 1);   /* abbrev entry number  */
+
+        saa_write32(p_info, producer_str_offset); /* offset from string table for DW_AT_producer  */
+        saa_write16(p_info, DW_LANG_Mips_Assembler); /* DW_AT_language  */
+        saa_write32(p_info, module_str_offset); /* offset from string table for DW_AT_name  */
+        saa_write32(p_info, dir_str_offset); /* offset from string table for DW_AT_comp_dir */
+        saa_write32(p_info, 0); /* DW_AT_stmt_list  */
+
+        if (ofmt == &of_macho64) {
+            saa_write64(p_info, 0); /* DW_AT_low_pc */
+            saa_write64(p_info, high_addr); /* DW_AT_high_pc */
+        } else  {
+            saa_write32(p_info, 0); /* DW_AT_low_pc */
+            saa_write32(p_info, high_addr); /* DW_AT_high_pc */
+        }
+
+        saa_write8(p_info, 2);   /* abbrev entry number */
+
+        if (ofmt == &of_macho64) {
+            saa_write64(p_info, 0); /* DW_AT_low_pc */
+            saa_write64(p_info, 0); /* DW_AT_frame_base */
+        } else  {
+            saa_write32(p_info, 0); /* DW_AT_low_pc */
+            saa_write32(p_info, 0); /* DW_AT_frame_base */
+        }
+        saa_write8(p_info, DW_END_default);
+
+        saa_len = p_info->datalen;
+        p_buf_base = p_buf = nasm_malloc(saa_len + 4); /* 4B for size info */
+
+        WRITELONG(p_buf, saa_len);
+        saa_rnbytes(p_info, p_buf, saa_len);
+        macho_output(p_section->index, p_buf_base, OUT_RAWDATA, saa_len + 4, NO_SEG, 0);
+
+        saa_free(p_info);
+        nasm_free(p_buf_base);
+    }
+
+    /* abbrev section */
+    {
+        struct SAA *p_abbrev = saa_init(1L);
+
+        p_section = get_section_by_name("__DWARF", "__debug_abbrev");
+        nasm_assert(p_section != NULL);
+
+        saa_write8(p_abbrev, 1); /* entry number */
+
+        saa_write8(p_abbrev, DW_TAG_compile_unit);
+        saa_write8(p_abbrev, DW_CHILDREN_yes);
+
+        saa_write8(p_abbrev, DW_AT_producer);
+        saa_write8(p_abbrev, DW_FORM_strp);
+
+        saa_write8(p_abbrev, DW_AT_language);
+        saa_write8(p_abbrev, DW_FORM_data2);
+
+        saa_write8(p_abbrev, DW_AT_name);
+        saa_write8(p_abbrev, DW_FORM_strp);
+
+        saa_write8(p_abbrev, DW_AT_comp_dir);
+        saa_write8(p_abbrev, DW_FORM_strp);
+
+        saa_write8(p_abbrev, DW_AT_stmt_list);
+        saa_write8(p_abbrev, DW_FORM_data4);
+
+        saa_write8(p_abbrev, DW_AT_low_pc);
+        saa_write8(p_abbrev, DW_FORM_addr);
+
+        saa_write8(p_abbrev, DW_AT_high_pc);
+        saa_write8(p_abbrev, DW_FORM_addr);
+
+        saa_write16(p_abbrev, DW_END_default);
+
+        saa_write8(p_abbrev, 2); /* entry number */
+
+        saa_write8(p_abbrev, DW_TAG_subprogram);
+        saa_write8(p_abbrev, DW_CHILDREN_no);
+
+        saa_write8(p_abbrev, DW_AT_low_pc);
+        saa_write8(p_abbrev, DW_FORM_addr);
+
+        saa_write8(p_abbrev, DW_AT_frame_base);
+        saa_write8(p_abbrev, DW_FORM_addr);
+
+        saa_write16(p_abbrev, DW_END_default);
+
+        saa_len = p_abbrev->datalen;
+
+        p_buf = nasm_malloc(saa_len);
+
+        saa_rnbytes(p_abbrev, p_buf, saa_len);
+        macho_output(p_section->index, p_buf, OUT_RAWDATA, saa_len, NO_SEG, 0);
+
+        saa_free(p_abbrev);
+        nasm_free(p_buf);
+    }
+}
+
+static void new_file_list (const char *file_name, const char *dir_name)
+{
+    struct dir_list *dir_list;
+    bool need_new_dir_list = true;
+
+    nasm_new(dw_cur_file);
+    dw_cur_file->file = ++dw_num_files;
+    dw_cur_file->file_name = file_name;
+    if(!dw_head_file) {
+        dw_head_file = dw_cur_file;
+    } else {
+        *dw_last_file_next = dw_cur_file;
+    }
+    dw_last_file_next = &(dw_cur_file->next);
+
+    if(dw_head_dir) {
+        list_for_each(dir_list, dw_head_dir) {
+            if(!(strcmp(dir_name, dir_list->dir_name))) {
+                dw_cur_file->dir = dir_list;
+                need_new_dir_list = false;
+                break;
+            }
+        }
+    }
+
+    if(need_new_dir_list)
+    {
+        nasm_new(dir_list);
+        dir_list->dir = dw_num_dirs++;
+        dir_list->dir_name = dir_name;
+        if(!dw_head_dir) {
+            dw_head_dir = dir_list;
+        } else {
+            *dw_last_dir_next = dir_list;
+        }
+        dw_last_dir_next = &(dir_list->next);
+        dw_cur_file->dir = dir_list;
+    }
+}
+
+static void macho_dbg_init(void)
+{
+}
+
+static void macho_dbg_linenum(const char *file_name, int32_t line_num, int32_t segto)
+{
+    bool need_new_list = true;
+    const char *cur_file = nasm_basename(file_name);
+    const char *cur_dir  = nasm_dirname(file_name);
+    (void)segto;
+
+    if(!dw_cur_file || strcmp(cur_file, dw_cur_file->file_name) ||
+        strcmp(cur_dir, dw_cur_file->dir->dir_name)) {
+        if(dw_head_file) {
+            struct file_list *match;
+
+            list_for_each(match, dw_head_file) {
+                if(!(strcmp(cur_file, match->file_name)) &&
+                    !(strcmp(cur_dir, match->dir->dir_name))) {
+                    dw_cur_file = match;
+                    dw_cur_file->dir = match->dir;
+                    need_new_list = false;
+                    break;
+                }
+            }
+        }
+
+        if(need_new_list) {
+            new_file_list(cur_file, cur_dir);
+        }
+    }
+
+    dbg_immcall = true;
+    cur_line = line_num;
+}
+
+static void macho_dbg_output(int type, void *param)
+{
+    struct section_info *sinfo_param = (struct section_info *)param;
+    int32_t secto = sinfo_param->secto;
+    bool need_new_sect = false;
+    struct SAA *p_linep = NULL;
+    (void)type;
+
+    if(!(dw_cur_sect && (dw_cur_sect->section == secto))) {
+        need_new_sect = true;
+        if(dw_head_sect) {
+            struct dw_sect_list *match = dw_head_sect;
+            uint32_t idx = 0;
+
+            for(; idx < dw_num_sects; idx++) {
+                if(match->section == secto) {
+                    dw_cur_sect = match;
+                    need_new_sect = false;
+                    break;
+                }
+                match = match->next;
+            }
+        }
+    }
+
+    if(need_new_sect) {
+        nasm_new(dw_cur_sect);
+        dw_num_sects ++;
+        p_linep = dw_cur_sect->psaa = saa_init(1L);
+        dw_cur_sect->line = dw_cur_sect->file = 1;
+        dw_cur_sect->offset = 0;
+        dw_cur_sect->next = NULL;
+        dw_cur_sect->section = secto;
+
+        saa_write8(p_linep, DW_LNS_extended_op);
+        saa_write8(p_linep, (ofmt == &of_macho64) ? 9 : 5);
+        saa_write8(p_linep, DW_LNE_set_address);
+        if (ofmt == &of_macho64) {
+            saa_write64(p_linep, 0);
+        } else {
+            saa_write32(p_linep, 0);
+        }
+
+        if(!dw_head_sect) {
+            dw_head_sect = dw_last_sect = dw_cur_sect;
+        } else {
+            dw_last_sect->next = dw_cur_sect;
+            dw_last_sect = dw_cur_sect;
+        }
+    }
+
+    if(dbg_immcall == true) {
+        int32_t line_delta = cur_line - dw_cur_sect->line;
+        int32_t offset_delta = sinfo_param->size - dw_cur_sect->offset;
+        uint32_t cur_file = dw_cur_file->file;
+        p_linep = dw_cur_sect->psaa;
+
+        if(cur_file != dw_cur_sect->file) {
+            saa_write8(p_linep, DW_LNS_set_file);
+            saa_write8(p_linep, cur_file);
+            dw_cur_sect->file = cur_file;
+        }
+
+        if(line_delta) {
+            int special_opcode = (line_delta - DW_LN_BASE) + (DW_LN_RANGE * offset_delta) +
+                                             DW_OPCODE_BASE;
+
+            if((line_delta >= DW_LN_BASE) && (line_delta < DW_MAX_LN) &&
+                (special_opcode < DW_MAX_SP_OPCODE)) {
+                saa_write8(p_linep, special_opcode);
+            } else {
+                saa_write8(p_linep, DW_LNS_advance_line);
+                saa_wleb128s(p_linep, line_delta);
+                if(offset_delta) {
+                    saa_write8(p_linep, DW_LNS_advance_pc);
+                    saa_wleb128u(p_linep, offset_delta);
+                }
+                saa_write8(p_linep, DW_LNS_copy);
+            }
+
+            dw_cur_sect->line = cur_line;
+            dw_cur_sect->offset = sinfo_param->size;
+        }
+
+        dbg_immcall = false;
+    }
+}
+
+static void macho_dbg_cleanup(void)
+{
+    /* dwarf sectors generation */
+    macho_dbg_generate();
+
+    {
+        struct dw_sect_list *p_sect = dw_head_sect;
+        struct file_list *p_file = dw_head_file;
+        uint32_t idx = 0;
+
+        for(; idx < dw_num_sects; idx++) {
+            struct dw_sect_list *next = p_sect->next;
+            nasm_free(p_sect);
+            p_sect = next;
+        }
+
+        for(idx = 0; idx < dw_num_files; idx++) {
+            struct file_list *next = p_file->next;
+            nasm_free(p_file);
+            p_file = next;
+        }
+    }
+}
+
 #ifdef OF_MACHO32
 static const struct macho_fmt macho32_fmt = {
     4,
@@ -1618,16 +2243,32 @@ static void macho32_init(void)
     macho_gotpcrel_sect = NO_SEG;
 }
 
-struct ofmt of_macho32 = {
+static const struct dfmt macho32_df_dwarf = {
+    "MachO32 (i386) dwarf debug format for Darwin/MacOS",
+    "dwarf",
+    macho_dbg_init,
+    macho_dbg_linenum,
+    null_debug_deflabel,
+    null_debug_directive,
+    null_debug_typevalue,
+    macho_dbg_output,
+    macho_dbg_cleanup,
+    NULL /*pragma list*/
+};
+
+static const struct dfmt * const macho32_df_arr[2] =
+ { &macho32_df_dwarf, NULL };
+
+const struct ofmt of_macho32 = {
     "NeXTstep/OpenStep/Rhapsody/Darwin/MacOS X (i386) object files",
     "macho32",
     0,
     32,
-    null_debug_arr,
-    &null_debug_form,
+    macho32_df_arr,
+    &macho32_df_dwarf,
     macho_stdmac,
     macho32_init,
-    null_setinfo,
+    nasm_do_legacy_output,
     macho_output,
     macho_symdef,
     macho_section,
@@ -1635,7 +2276,8 @@ struct ofmt of_macho32 = {
     macho_segbase,
     null_directive,
     macho_filename,
-    macho_cleanup
+    macho_cleanup,
+    macho_pragma_list,
 };
 #endif
 
@@ -1665,16 +2307,32 @@ static void macho64_init(void)
     define_label("..gotpcrel", macho_gotpcrel_sect, 0L, NULL, false, false);
 }
 
-struct ofmt of_macho64 = {
+static const struct dfmt macho64_df_dwarf = {
+    "MachO64 (x86-64) dwarf debug format for Darwin/MacOS",
+    "dwarf",
+    macho_dbg_init,
+    macho_dbg_linenum,
+    null_debug_deflabel,
+    null_debug_directive,
+    null_debug_typevalue,
+    macho_dbg_output,
+    macho_dbg_cleanup,
+    NULL /*pragma list*/
+};
+
+static const struct dfmt * const macho64_df_arr[2] =
+ { &macho64_df_dwarf, NULL };
+
+const struct ofmt of_macho64 = {
     "NeXTstep/OpenStep/Rhapsody/Darwin/MacOS X (x86_64) object files",
     "macho64",
     0,
     64,
-    null_debug_arr,
-    &null_debug_form,
+    macho64_df_arr,
+    &macho64_df_dwarf,
     macho_stdmac,
     macho64_init,
-    null_setinfo,
+    nasm_do_legacy_output,
     macho_output,
     macho_symdef,
     macho_section,
@@ -1682,7 +2340,8 @@ struct ofmt of_macho64 = {
     macho_segbase,
     null_directive,
     macho_filename,
-    macho_cleanup
+    macho_cleanup,
+    macho_pragma_list,
 };
 #endif
 
